//---------------------------------------------------------------------------

#pragma hdrstop

#include "FormSettings.h"

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <assert.h>
#include "Scope.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream>
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;
CALLBACK_UPDATE_TIMEBASE lpUpdateTimebaseFn = NULL;
CALLBACK_UPDATE_SENSITIVITY lpUpdateSensitivityFn = NULL;
CALLBACK_UPDATE_COUPLING lpUpdateCouplingFn = NULL;


Scope *scope = Scope::Instance();

void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void UpdateTimebase(int newIndex)
{
	if (lpUpdateTimebaseFn)
		lpUpdateTimebaseFn(callbackCookie, newIndex);
}

void UpdateSensitivity(int channel, int newIndex)
{
	if (lpUpdateSensitivityFn)
		lpUpdateSensitivityFn(callbackCookie, channel, newIndex);
}

void UpdateCoupling(int channel, enum E_COUPLING_TYPE type)
{
	if (lpUpdateCouplingFn)
		lpUpdateCouplingFn(callbackCookie, channel, type);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "dso138.dll loaded");

	scope->callbackLog = Log;
	scope->callbackData = Data;
	scope->callbackConnect = Connect;
	scope->callbackTrigger = Trigger;
	scope->callbackUpdateTimebase = UpdateTimebase;
	scope->callbackUpdateSensitivity = UpdateSensitivity;
	scope->callbackUpdateCoupling = UpdateCoupling;
}

static enum E_TRIGGER_MODE eMode[] = { TRIGGER_MODE_AUTO, TRIGGER_MODE_NORMAL, TRIGGER_MODE_SINGLE, TRIGGER_MODE_CONTINUOUS };
static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_GND, COUPLING_AC, COUPLING_DC };
unsigned int iBuf[] = { Scope::SAMPLES_COUNT };
static struct S_SCOPE_CAPABILITIES capabilities =
{
	Scope::CHANNELS,
	12,		//bits/sample
	sizeof(Scope::fSens)/sizeof(Scope::fSens[0]), Scope::fSens,
	sizeof(Scope::fSampl)/sizeof(Scope::fSampl[0]), Scope::fSampl,
	sizeof(eMode)/sizeof(eMode[0]), eMode,
	sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
	sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
	sizeof(iBuf)/sizeof(iBuf[0]), iBuf
};

bool pChannelEnable[Scope::CHANNELS];
unsigned int pSensitivity[Scope::CHANNELS];
enum E_COUPLING_TYPE eCouplingType[Scope::CHANNELS];

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[Scope::CHANNELS][sizeof(Scope::fSens)/sizeof(Scope::fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	frmSettings = new TfrmSettings(NULL);
	frmSettings->comPort = scope->comPort.c_str();
	frmSettings->ShowModal();
	if (frmSettings->confirmed)
	{
		scope->comPort = frmSettings->comPort.c_str();
	}
	delete frmSettings;
}

int __stdcall EnableChannel(int iChannelId, bool state)
{
	if (scope == NULL)
		return -1;
	return scope->EnableChannel(iChannelId, state);
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(Scope::fSens)/sizeof(Scope::fSens[0])))
		return -1;
	return 0;
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(Scope::fSampl)/sizeof(Scope::fSampl[0])))
		return -1;
	return 0;
}

#if 0
int __stdcall SetTriggerSource(int channel, int iId)
{
	if (channel < 0 || channel >= Scope::CHANNELS)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -2;
	return 0;
}

int __stdcall SetTriggerLevel(int level)
{
	return 0;
}

int __stdcall SetTriggerPosition(int pos)
{
	return 0;
}
#endif

int __stdcall SetCouplingType(int iChannel, int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eCoupling)/sizeof(eCoupling[0])))
		return -1;
	return 0;
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return 0;
}

int __stdcall Connect(void)
{
	return scope->Connect();
}

int __stdcall Disconnect(void)
{
	return scope->Disconnect();
}

int __stdcall Run(bool state)
{
	return scope->Run(state);
}

int __stdcall SetTriggerMode(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eMode)/sizeof(eMode[0])))
		return -1;
	return 0;
}

#if 0
int __stdcall ManualTrigger(void)
{
	return scope->ManualTrigger();
}
#endif

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;
	
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		settings->pChannelEnable[i] = true;
		settings->pSensitivity[i] = 0;
		settings->eCouplingType[i] = COUPLING_DC;
	}

	settings->iSamplingPeriod = 0;
	settings->iTriggerSrcChannel = 0;
	settings->eTriggerMode = TRIGGER_MODE_AUTO;
	settings->eTriggerType = TRIGGER_SLOPE_UP;
	settings->iTriggerLevel = 0;
	settings->iBufferSize = 512;
	for (int i=0; i<Scope::CHANNELS; i++)
	{
		for (unsigned int j=0; j<sizeof(calibration)/sizeof(calibration[0]); j++)
		{
			calibration[i][j].offset = 0;
			calibration[i][j].gain = 1.0;
		}
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    settings->iChannelCount = Scope::CHANNELS;
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		settings->pChannelEnable[i] = root["ChannelEnable"][i].asBool();
		settings->pSensitivity[i] = root["Sensitivity"][i].asInt();
		if (settings->pSensitivity[i] >= sizeof(Scope::fSens)/sizeof(Scope::fSens[0]))
			settings->pSensitivity[i] = 0;

		int coupling = root["CouplingType"][i].asInt();
		if (Utils::in_group(coupling, eCoupling))
			settings->eCouplingType[i] = (E_COUPLING_TYPE)coupling;
		else
			settings->eCouplingType[i] = COUPLING_DC;
	}

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(Scope::fSampl)/sizeof(Scope::fSampl[0]))
		settings->iSamplingPeriod = 0;

	int mode = root.get("TriggerMode", TRIGGER_MODE_AUTO).asInt();
	if (Utils::in_group(mode, eMode))
		settings->eTriggerMode = (E_TRIGGER_MODE)mode;
	else
		settings->eTriggerMode = TRIGGER_MODE_AUTO;

	int trigger = root.get("TriggerType", TRIGGER_SLOPE_UP).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_SLOPE_UP;

	settings->iTriggerSrcChannel = 0;
	
	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	settings->iTriggerPosition = root.get("TriggerPosition", 50).asInt();
	if (settings->iTriggerPosition < 0 || settings->iTriggerPosition > 100)
		settings->iTriggerPosition = 50;

	settings->iBufferSize = root.get("BufferSize", 512).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	// read calibration data also
	const Json::Value rootcal = root["Calibration"];	
	for (int i=0; i<Scope::CHANNELS; i++)
	{
		const Json::Value ch_cal = rootcal[i];
		if (ch_cal.type() != Json::arrayValue)
		{
			for (unsigned int j=0; j<sizeof(calibration[i])/sizeof(calibration[i][0]); j++)
			{
				calibration[i][j].offset = 0;
				calibration[i][j].gain = 1.0f;
			}
		}
		else
		{
			for (unsigned int j=0; j<sizeof(calibration[i])/sizeof(calibration[i][0]); j++)
			{
				calibration[i][j].offset = rootcal[i][j].get("Offset", 0).asInt();
				calibration[i][j].gain = rootcal[i][j].get("Gain", 1.0).asDouble();
			}
		}
	}

	root.getString("comPort", scope->comPort);

	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
	Json::StyledWriter writer;
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		root["ChannelEnable"][i] = settings->pChannelEnable[i];
		root["Sensitivity"][i] = settings->pSensitivity[i];
		root["CouplingType"][i] = settings->eCouplingType[i];
	}
	root["SamplingPeriod"] = settings->iSamplingPeriod;
    root["TriggerMode"] = settings->eTriggerMode;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	root["TriggerPosition"] = settings->iTriggerPosition;
	root["BufferSize"] = settings->iBufferSize;

	// add calibration data also
	const Json::Value rootcal = root["Calibration"];
	for (unsigned int i=0; i<capabilities.iChannelCount; i++)
	{
		for (unsigned int j=0; j<sizeof(calibration)/sizeof(calibration[0]); j++)
		{
			  root["Calibration"][i][j]["Offset"] = calibration[i][j].offset;
			  root["Calibration"][i][j]["Gain"] = calibration[i][j].gain;
		}
	}

	root["comPort"] = scope->comPort;

	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int channel, int range, int offset, float gain)
{
	if (channel < 0 || channel >= Scope::CHANNELS)
		return -1;
	if (range < 0 || range >= (int)(sizeof(Scope::fSens)/sizeof(Scope::fSens[0])))
		return -2;
	calibration[channel][range].offset = offset;
	calibration[channel][range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int channel, int range, int* const offset, float* const gain)
{
	if (channel < 0 || channel >= Scope::CHANNELS)
		return -1;
	if (range < 0 || range >= (int)(sizeof(Scope::fSens)/sizeof(Scope::fSens[0])))
		return -2;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[channel][range].offset;
	*gain = calibration[channel][range].gain;
	return 0;
}

int __stdcall SetUpdateTimebaseCallback(CALLBACK_UPDATE_TIMEBASE f)
{
	lpUpdateTimebaseFn = f;
	return 0;
}

int __stdcall SetUpdateSensitivityCallback(CALLBACK_UPDATE_SENSITIVITY f)
{
	lpUpdateSensitivityFn = f;
	return 0;
}

int __stdcall SetUpdateCouplingCallback(CALLBACK_UPDATE_COUPLING f)
{
	lpUpdateCouplingFn = f;
	return 0;
}

